from fastapi import BackgroundTasks, FastAPI, HTTPException, Request
import fastapi
import fastapi.responses

import datetime
import logging
import os
import random
import string
import traceback
from typing import List, Dict

from uploader.config import BasicSettings
from uploader.errors import ActiveWorkflowException, InternalException
from uploader import models_activeworkflow as awmodels

from uploader.utils import (
    retrieve_files_by_mask,
    retrieve_files_by_name,
    retrieve_options_from_parameters,
)
from uploader import store

app = FastAPI()
# TODO: this is state, remove it!!
running_uploads: Dict[str, dict] = {}
completed_uploads: Dict[str, awmodels.ResponseCheck] = {}


@app.get("/appinfo")
def appinfo():
    """ Prints a part of the configuration of this datalake-agent instance.
    """

    print(BasicSettings())
    return BasicSettings().dict(exclude={"cdstar_pass", "cdstar_user"})


@app.post("/datalake", response_model=awmodels.ResponseCommon)
def datalake(payload: awmodels.RequestCommon, background_tasks: BackgroundTasks):
    """ Implements an remote agent API endpoint as specified by active_workflow, see 
    [active_workflow docs](https://github.com/automaticmode/active_workflow/blob/master/docs/remote_agent_api.md) for more details. 
    See the explicit API endpoints for each method to find more information on them.
    """

    if payload.method is not None and payload.method == "register":
        return register(
            awmodels.RequestRegister.parse_obj(payload.dict()), background_tasks
        )
    elif payload.method is not None and payload.method == "receive":
        return receive(
            awmodels.RequestReceive.parse_obj(payload.dict()), background_tasks
        )
    elif payload.method is not None and payload.method == "check":
        return check(awmodels.RequestCheck.parse_obj(payload.dict()), background_tasks)

    raise HTTPException(
        status_code=400,
        detail="You need to supply a valid method: register, receive, check.",
    )


@app.post("/datalake/register", response_model=awmodels.ResponseRegister)
def register(payload: awmodels.RequestRegister, background_tasks: BackgroundTasks):
    """ Handle register method of active_workflow. Check 
    [active_workflow docs](https://github.com/automaticmode/active_workflow/blob/master/docs/remote_agent_api.md#register-method) for 
    more information on the method itself. The response from this method will provide some examples for annotations that may be sent.
    """

    logging.info(f"Register - {datetime.datetime.now().isoformat()} - {payload.dict()}")

    resp = awmodels.ResponseRegister()

    return fastapi.responses.JSONResponse(
        status_code=fastapi.status.HTTP_200_OK, content=resp.dict()
    )


@app.post("/datalake/receive", response_model=awmodels.ResponseReceive)
def receive(payload: awmodels.RequestReceive, background_tasks: BackgroundTasks):
    """ Starts and upload and annotation of a list of files.
    
    Handle the receive method of ActiveWorkflow, see
    [active_workflow docs](https://github.com/automaticmode/active_workflow/blob/master/docs/remote_agent_api.md#receive-method) for more information.
    This method checks `$.params.message.payload.files_moved` and expects a list of filenames there. It will iterate through each of these filenames
    and upload them to a new CDSTAR Archive per message. The filenames are expected to reside under
    `base_directory`/`source_directory` as configured (or overriden in the agent options in active workflow.)
    Metadata annotations supplied under `$.params.message.payload.annotations` or 
    (with lower priority) under `$.params.options.annotations` are added to generated metadata (from CDSTAR responses)
    and uploaded to CouchDB. Annotations starting with `@` will be ignored, they will be set by the application where applicable.
    """

    logging.info(f"Receive - {datetime.datetime.now().isoformat()} - {payload.dict()}")

    msg_payload = payload.params.message.payload
    if msg_payload.files_moved is None or len(msg_payload.files_moved) < 1:
        resp = awmodels.ResponseReceive()
        resp.result.memory = payload.params.memory
        return fastapi.responses.JSONResponse(status_code=200, content=resp.dict(),)

    opts = retrieve_options_from_parameters(payload.params)

    filelist = retrieve_files_by_name(
        subdir=opts["source_directory"], namelist=msg_payload.files_moved
    )

    upload_id = "".join(random.choices(string.ascii_letters, k=12))
    while upload_id in list(running_uploads.keys()) + list(completed_uploads.keys()):
        upload_id = "".join(random.choices(string.ascii_letters, k=12))

    running_uploads[upload_id] = {"started": datetime.datetime.now()}
    background_tasks.add_task(
        perform_datalake_upload, filelist=filelist, options=opts, upload_id=upload_id,
    )

    resp = awmodels.ResponseReceive()
    resp.result.logs.append(f"The upload has been started with id {upload_id}")
    resp.result.memory.uploads.append(upload_id)
    return fastapi.responses.JSONResponse(
        status_code=fastapi.status.HTTP_202_ACCEPTED, content=resp.dict(),
    )


@app.post("/datalake/check", response_model=awmodels.ResponseCheck)
def check(payload: awmodels.RequestCheck, background_tasks: BackgroundTasks):
    """ Check on data uploads as supplied by `$.params.memory.uploads`.
    
    Handles the check method of ActiveWorkflow, see 
    [active_workflow docs](https://github.com/automaticmode/active_workflow/blob/master/docs/remote_agent_api.md#check-method) for more information.
    This method uses upload ids supplied by the memory part of a request to check on all uploads. An upload may be running or completed,
    which will be reported in the logs or message section respectively. Invalid / unknown upload ids are removed from the memory.
    """

    logging.info(f"Check - {datetime.datetime.now().isoformat()} - {payload.dict()}")

    response = awmodels.ResponseCheck()
    if len(payload.params.memory.uploads) < 1:
        return fastapi.responses.JSONResponse(
            status_code=fastapi.status.HTTP_200_OK, content=response.dict(),
        )

    for upl in payload.params.memory.uploads:
        if upl in running_uploads.keys():
            response.result.memory.uploads.append(upl)
            response.result.logs.append(
                f"Upload {upl} is running since {running_uploads[upl]['started'].isoformat()}"
            )
        elif upl in completed_uploads.copy().keys():
            finished_upl = completed_uploads.pop(upl)
            response.result.logs = response.result.logs + finished_upl.result.logs
            response.result.errors = response.result.errors + finished_upl.result.errors
            response.result.messages = (
                response.result.messages + finished_upl.result.messages
            )
        else:
            logging.warning(f"Found orphan upload id: {upl}")

    return fastapi.responses.JSONResponse(
        status_code=fastapi.status.HTTP_200_OK, content=response.dict()
    )


def perform_datalake_upload(filelist: List[str], options: dict, upload_id: str) -> None:
    """ Perform the upload and annotation within the datalake.
    From the options and annotations supplied by the caller or configured in the application, find all files
    using `source_directory` and `source_filemask` from the `options` parameter. Upload each file to a new
    archive created in CDSTAR and create metadata annotations supplied by the `annotations` parameter. Each key
    in annotations will be filtered against the internal models of `models_metadata`.

    Parameters
    ----------
    options : dict
        A dictionary of options that are supplied by the caller. Specifically, `source_directory`, `source_filemask` and 
        `delete_after_upload` have to be set.
    annotations : dict
        A dictionary of annotations that are added to the CouchDB metadata after the files have been uploaded to CDSTAR.
    """
    resp = awmodels.ResponseCheck()

    if filelist is None or len(filelist) < 1:
        resp.result.logs.append(f"No files found to upload for id {upload_id}.")
        upload_completed(upload_id, resp)
        return

    # upload all files to a new cdstar archive
    try:
        uploaded = store.upload_files(filelist, mime_type=options["mime_type"])
        resp.result.messages.append(
            awmodels.MessageOutput(upload_id=upload_id, archive_id=uploaded.id)
        )
        resp.result.logs.append(
            f"Upload to CDSTAR for id {upload_id} finished at {datetime.datetime.now().isoformat()}."
        )
    except:
        logging.error(traceback.format_exc())
        resp.result.errors.append(
            f"I could not establish a connection to CDSTAR in some way. Upload with id {upload_id} failed."
        )
        upload_completed(upload_id, resp)
        return

    resp.result.logs.append(
        uploaded.json(exclude_none=True, exclude_defaults=True, exclude_unset=True)
    )
    for file in uploaded.files:
        if file.id is not None:
            if options["delete_after_upload"]:
                os.remove(
                    os.path.join(
                        BasicSettings().base_directory,
                        options["source_directory"].lstrip("/"),
                        file.name.lstrip("/"),
                    )
                )
        else:
            resp.result.errors.append(f"Error: {file.name} could not be uploaded.")

    upload_completed(upload_id, resp)


def upload_completed(upload_id: str, response: awmodels.ResponseCheck):
    upl = running_uploads.pop(upload_id, None)
    if upl is None:
        response.result.errors.append(
            f"Upload ID {upload_id} is unknown or was not started properly before it has been finished."
        )
    completed_uploads[upload_id] = response
    logging.info(f"Completed upload of {upload_id}.")


@app.exception_handler(ActiveWorkflowException)
async def aw_exception_handler(request: Request, exc: ActiveWorkflowException):
    logging.error(f"{exc.name}: {exc.message}")
    resp = awmodels.ResponseCheck()
    resp.result.errors.append(f"{exc.name}: {exc.message}")

    return fastapi.responses.JSONResponse(
        status_code=fastapi.status.HTTP_400_BAD_REQUEST, content=resp.dict(),
    )


@app.exception_handler(InternalException)
async def int_exception_handler(request: Request, exc: InternalException):
    logging.error(f"{exc.name}: {exc.message}")
    resp = awmodels.ResponseCheck()
    resp.result.errors.append(f"{exc.name}: {exc.message}")

    return fastapi.responses.JSONResponse(
        status_code=fastapi.status.HTTP_500_INTERNAL_SERVER_ERROR, content=resp.dict(),
    )

