import abc
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Field, validator

from uploader import config

#
# Common models
#


class ParamsCommon(BaseModel, abc.ABC):
    """
    These are all possible parameters that may be supplied using the Active Workflow Remote Agent API.
    """

    message: Optional[Dict[str, Any]] = {}
    options: Optional[Dict[str, Any]] = {}
    memory: Optional[Dict[str, Any]] = {}
    credentials: Optional[Dict[str, Any]] = {}


class ResultCommon(BaseModel, abc.ABC):
    """
    These are all possible result values that may be returned using the Active Workflow Remote Agent API.
    """

    errors: Optional[List[str]] = []
    logs: Optional[List[str]] = []
    memory: Optional[Dict[str, Any]] = {}
    messages: Optional[List[Dict[str, Any]]] = []
    name: Optional[str] = ""
    display_name: Optional[str] = ""
    description: Optional[str] = ""
    default_options: Optional[Dict[str, Any]] = {}


class RequestCommon(BaseModel, abc.ABC):
    """
    These are all possible values in a request from Active Workflow using the Remote Agent API.
    """

    method: str = Field(..., example="receive")
    params: ParamsCommon = Field(..., example=ParamsCommon())

    @validator("method")
    def method_choices(cls, v):
        if v not in ["register", "receive", "check"]:
            raise ValueError(
                "Method has to be one of 'register', 'receive' or 'check'."
            )
        return v


class ResponseCommon(BaseModel, abc.ABC):
    """
    This is a possible reponse returned when using the Acitve Workflow Remote Agent API.
    """

    result: ResultCommon = Field(..., example=ResultCommon())


#
# Application specifics
#


class PayloadInput(BaseModel):
    """
    This model represents the expected input payload when a message is received via `receive`.
    """

    files_moved: List[str] = Field([], example=["test.csv", "test2.csv"])
    options: Dict[str, Any] = Field({}, example={"mime_type": "text/plain"})

    @staticmethod
    def example() -> dict:
        return {
            "files_moved": ["test.csv", "test2.csv"],
            "options": {"mime_type": "text/plain"},
        }


class MessageOutput(BaseModel):
    """
    This model represents the produced output schema that will be provided in the messages.
    """

    upload_id: str = Field(..., example="DoamRsPwHzrl")
    archive_id: str = Field(..., example="iw7s2px9")

    @staticmethod
    def example() -> dict:
        return {"upload_id": "DoamRsPwHzrl", "archive_id": "iw7s2px9"}


class MemoryCommon(BaseModel):
    """
    This model represents the expected memory content to communicate state of the agent.
    """

    uploads: List[str] = Field([], example=["DoamRsPwHzrl", "lsZHWpxwMqaj"])

    @staticmethod
    def example() -> dict:
        return {"uploads": ["DoamRsPwHzrl", "lsZHWpxwMqaj"]}


#
# Register Models
#


class ParamsRegister(ParamsCommon):
    pass


class RequestRegister(RequestCommon):
    method: str = Field("register", const="register")
    params: ParamsRegister = Field(ParamsRegister(), const=ParamsRegister())


class ResultRegister(ResultCommon):
    name: str = config.BasicSettings().application_name.replace(" ", "")
    display_name: str = config.BasicSettings().application_name
    description: str = config.agent_description
    default_options: Dict[str, Any] = config.BasicSettings().dict(
        include=config.agent_config_parameters
    )


class ResponseRegister(ResponseCommon):
    result: ResultRegister = ResultRegister()


#
# Receive Models
#


class MessageInputReceive(BaseModel):
    payload: PayloadInput


class ParamsReceive(ParamsCommon):
    message: MessageInputReceive = Field(
        ..., example=MessageInputReceive(payload=PayloadInput())
    )
    options: Dict[str, Any] = Field(
        {}, example=config.BasicSettings().dict(include=config.agent_config_parameters),
    )
    memory: MemoryCommon = Field(
        MemoryCommon(), example=MemoryCommon(**MemoryCommon.example())
    )
    credentials: dict = Field({}, example={}, const={})


class RequestReceive(RequestCommon):
    method: str = Field("receive", const="receive")
    params: ParamsReceive = Field(
        ParamsReceive(
            message=MessageInputReceive(payload=PayloadInput(**PayloadInput.example()))
        ),
        example=ParamsReceive(message=MessageInputReceive(payload=PayloadInput())),
    )


class ResultReceive(ResultCommon):
    errors: List[str] = Field([], example=[])
    logs: List[str] = Field([], example=[])
    memory: MemoryCommon = Field(
        MemoryCommon(), example=MemoryCommon(**MemoryCommon.example())
    )
    messages: List[MessageOutput] = Field(
        [], example=[MessageOutput(**MessageOutput.example())]
    )


class ResponseReceive(ResponseCommon):
    result: ResultReceive = Field(ResultReceive(), example=ResultReceive())


#
# Check models
#


class ParamsCheck(ParamsCommon):
    message: Any = Field(None, const=None)
    options: Dict[str, Any] = Field(
        {}, example=config.BasicSettings().dict(include=config.agent_config_parameters),
    )
    memory: MemoryCommon = Field(
        MemoryCommon(), example=MemoryCommon(**MemoryCommon.example())
    )
    credentials: dict = Field({}, const={})


class RequestCheck(RequestCommon):
    method: str = Field("check", const="check")
    params: ParamsCheck = Field(ParamsCheck(), example=ParamsCheck())


class ResultCheck(ResultCommon):
    errors: List[str] = Field([], example=[])
    logs: List[str] = Field([], example=[])
    memory: MemoryCommon = Field(
        MemoryCommon(), example=MemoryCommon(**MemoryCommon.example())
    )
    messages: List[MessageOutput] = Field(
        [], example=[MessageOutput(**MessageOutput.example())]
    )


class ResponseCheck(ResponseCommon):
    result: ResultCheck = Field(ResultCheck(), example=ResultCheck())
